#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Methods and / or classes to perform colocation
"""
import numpy as np
import os
import pandas as pd
from pyaerocom import logger
from pyaerocom.exceptions import (ColocationError, 
                                  DataUnitError,
                                  DimensionOrderError,
                                  MetaDataError,
                                  TimeMatchError,
                                  VarNotAvailableError)

from pyaerocom.time_config import TS_TYPE_TO_PANDAS_FREQ
from pyaerocom.helpers import (to_pandas_timestamp, 
                               to_datestring_YYYYMMDD)

from pyaerocom.filter import Filter
from pyaerocom.colocateddata import ColocatedData
        
def colocate_gridded_gridded(gridded_data, gridded_data_ref, ts_type=None,
                             start=None, stop=None, filter_name=None, 
                             regrid_res_deg=None, remove_outliers=True,
                             vert_scheme=None, harmonise_units=True,
                             regrid_scheme='areaweighted', 
                             var_outlier_ranges=None,
                             var_ref_outlier_ranges=None,
                             update_baseyear_gridded=None,
                             apply_time_resampling_constraints=None,
                             min_num_obs=None,
                             colocate_time=False,
                             var_keep_outliers=True,
                             var_ref_keep_outliers=False, 
                             **kwargs):
    """Colocate 2 gridded data objects
    
    Todo
    ----
    - think about vertical dimension (vert_scheme input not used at the moment)
    
    Parameters
    ----------
    gridded_data : GriddedData
        gridded data (e.g. model results)
    gridded_data_ref : GriddedData
        reference dataset that is used to evaluate 
        :attr:`gridded_data` (e.g. gridded observation data)
    ts_type : str
        desired temporal resolution of colocated data (must be valid AeroCom
        ts_type str such as daily, monthly, yearly..)
    start : :obj:`str` or :obj:`datetime64` or similar, optional
        start time for colocation, if None, the start time of the input
        :class:`GriddedData` object is used
    stop : :obj:`str` or :obj:`datetime64` or similar, optional
        stop time for colocation, if None, the stop time of the input
        :class:`GriddedData` object is used
    filter_name : str
        string specifying filter used (cf. :class:`pyaerocom.filter.Filter` for
        details). If None, then it is set to 'WORLD-wMOUNTAINS', which 
        corresponds to no filtering (world with mountains). 
        Use WORLD-noMOUNTAINS to exclude mountain sites.
    regrid_res_deg : :obj:`int`, optional
        regrid resolution in degrees. If specified, the input gridded data 
        objects will be regridded in lon / lat dimension to the input 
        resolution. (BETA feature)
    remove_outliers : bool
        if True, outliers are removed from model and obs data before colocation, 
        else not.
    vert_scheme : str
        string specifying scheme used to reduce the dimensionality in case 
        input grid data contains vertical dimension. Example schemes are 
        `mean, surface, altitude`, for details see 
        :func:`GriddedData.to_time_series`.
    harmonise_units : bool
        if True, units are attempted to be harmonised (note: raises Exception
        if True and units cannot be harmonised).
    regrid_scheme : str
        iris scheme used for regridding (defaults to area weighted regridding)
    var_outlier_ranges : :obj:`dict`, optional
        dictionary specifying outlier ranges for dataset to be analysed
        (e.g. dict(od550aer = [-0.05, 10], ang4487aer=[0,4])). If None, then
        the pyaerocom default outlier ranges are used for the input variable.
        Defaults to None.
    var_ref_outlier_ranges : dict, optional
        like `var_outlier_ranges` but for reference dataset.
    update_baseyear_gridded : int, optional
        optional input that can be set in order to redefine the time dimension
        in the gridded data object to be analysed. E.g., if the data object 
        is a climatology (one year of data) that has set the base year of the
        time dimension to a value other than the specified input start / stop 
        time this may be used to update the time in order to make colocation 
        possible.
    apply_time_resampling_constraints : bool, optional
        if True, then time resampling constraints are applied as provided via 
        :attr:`min_num_obs` or if that one is unspecified, as defined in
        :attr:`pyaerocom.const.OBS_MIN_NUM_RESAMPLE`. If None, than 
        :attr:`pyaerocom.const.OBS_APPLY_TIME_RESAMPLE_CONSTRAINTS` is used
        (which defaults to True !!).
    min_num_obs : int or dict, optional
        minimum number of observations for resampling of time
    colocate_time : bool
        if True and if original time resolution of data is higher than desired
        time resolution (`ts_type`), then both datasets are colocated in time 
        *before* resampling to lower resolution. 
    var_keep_outliers : bool
        if True, then no outliers will be removed from dataset to be analysed, 
        even if `remove_outliers` is True. That is because for model evaluation
        often only outliers are supposed to be removed in the observations but
        not in the model.
    var_ref_keep_outliers : bool
        if True, then no outliers will be removed from the reference dataset, 
        even if `remove_outliers` is True.
    **kwargs
        additional keyword args (not used here, but included such that factory 
        class can handle different methods with different inputs)
    
    Returns
    -------
    ColocatedData
        instance of colocated data
        
    """
    if vert_scheme is not None:
        raise NotImplementedError('Input vert_scheme cannot yet be handled '
                                  'for gridded / gridded colocation...')
    if ts_type is None:
        ts_type = 'monthly'
    
    if var_outlier_ranges is None:
        var_outlier_ranges = {}
    if var_ref_outlier_ranges is None:
        var_ref_outlier_ranges = {}
    if filter_name is None:
        filter_name = 'WORLD-wMOUNTAINS'
    if gridded_data.var_info.has_unit:
        if harmonise_units and not gridded_data.units == gridded_data_ref.units:
            try:
                gridded_data_ref.convert_unit(gridded_data.units)
            except:
                raise DataUnitError('Failed to merge data unit of reference '
                                    'gridded data object ({}) to data unit '
                                    'of gridded data object ({})'
                                    .format(gridded_data.units, 
                                            gridded_data_ref.units))
    var, var_ref = gridded_data.var_name, gridded_data_ref.var_name
    if remove_outliers:
        low, high, low_ref, high_ref = None, None, None, None    
        if var in var_outlier_ranges:
            low, high = var_outlier_ranges[var]
        if var_ref in var_ref_outlier_ranges:
            low_ref, high_ref = var_ref_outlier_ranges[var_ref]
    
    if update_baseyear_gridded is not None:
        # update time dimension in gridded data
        gridded_data.base_year = update_baseyear_gridded
        
    # get start / stop of gridded data as pandas.Timestamp
    grid_start = to_pandas_timestamp(gridded_data.start)
    grid_stop = to_pandas_timestamp(gridded_data.stop)
    
    grid_start_ref = to_pandas_timestamp(gridded_data_ref.start)
    grid_stop_ref = to_pandas_timestamp(gridded_data_ref.stop)
    
    grid_ts_type = gridded_data.ts_type
    
    if start is None:
        start = grid_start
    else:
        start = to_pandas_timestamp(start)    
    if stop is None:
        stop = grid_stop
    else:
        stop = to_pandas_timestamp(stop)
    
    if grid_start_ref > start:
        start = grid_start_ref
    if grid_stop_ref < stop:
        stop = grid_stop_ref
    # check overlap
    if stop < grid_start or start > grid_stop:
        raise TimeMatchError('Input time range {}-{} does not '
                             'overlap with data range: {}-{}'
                             .format(start, stop, grid_start, grid_stop))
    gridded_data = gridded_data.crop(time_range=(start, stop))
    gridded_data_ref = gridded_data_ref.crop(time_range=(start, stop))
    
    if regrid_res_deg is not None:
        
        lons = gridded_data_ref.longitude.points
        lats = gridded_data_ref.latitude.points
        
        lons_new = np.arange(lons.min(), lons.max(), regrid_res_deg)
        lats_new = np.arange(lats.min(), lats.max(), regrid_res_deg) 
        
        gridded_data_ref = gridded_data_ref.interpolate(latitude=lats_new, 
                                                        longitude=lons_new)
        
    
        
    # get both objects in same time resolution
    if not colocate_time:
        gridded_data = gridded_data.resample_time(ts_type)
        gridded_data_ref = gridded_data_ref.resample_time(ts_type)
    
    # guess bounds (for area weighted regridding, which is the default)
    gridded_data._check_lonlat_bounds()
    gridded_data_ref._check_lonlat_bounds()
    
    # perform regridding
    gridded_data = gridded_data.regrid(gridded_data_ref, 
                                       scheme=regrid_scheme)
    
    # perform region extraction (if applicable)
    regfilter = Filter(name=filter_name)
    gridded_data = regfilter(gridded_data)
    gridded_data_ref = regfilter(gridded_data_ref)
    
    if not gridded_data.shape == gridded_data_ref.shape:
        raise ColocationError('Shape mismatch between two colocated data '
                               'arrays, please debug')
    files_ref = [os.path.basename(x) for x in gridded_data_ref.from_files]
    files = [os.path.basename(x) for x in gridded_data.from_files]
    
    
    meta = {'data_source'       :   [gridded_data_ref.data_id,
                                     gridded_data.data_id],
            'var_name'          :   [var_ref, var],
            'ts_type'           :   ts_type,
            'filter_name'       :   filter_name,
            'ts_type_src'       :   [gridded_data_ref.ts_type, grid_ts_type],
            'start_str'         :   to_datestring_YYYYMMDD(start),
            'stop_str'          :   to_datestring_YYYYMMDD(stop),
            'var_units'         :   [str(gridded_data_ref.units),
                                     str(gridded_data.units)],
            'vert_scheme'       :   vert_scheme,
            'data_level'        :   3,
            'revision_ref'      :   gridded_data_ref.data_revision,
            'from_files'        :   files,
            'from_files_ref'    :   files_ref,
            'colocate_time'     :   colocate_time,
            'apply_constraints' :   apply_time_resampling_constraints,
            'min_num_obs'       :   min_num_obs}
    
    meta.update(regfilter.to_dict())
    if remove_outliers:
        if not var_keep_outliers:
            gridded_data.remove_outliers(low, high)
        if not var_ref_keep_outliers:
            gridded_data_ref.remove_outliers(low_ref, high_ref)
            
    data = gridded_data.grid.data
    if isinstance(data, np.ma.core.MaskedArray):
        data = data.filled(np.nan)
    data_ref = gridded_data_ref.grid.data
    if isinstance(data_ref, np.ma.core.MaskedArray):
        data_ref = data_ref.filled(np.nan)
    arr = np.asarray((data_ref,
                      data))
    time = gridded_data.time_stamps().astype('datetime64[ns]')
    lats = gridded_data.latitude.points
    lons = gridded_data.longitude.points
    
    
    # create coordinates of DataArray
    coords = {'data_source' : meta['data_source'],
              'var_name'    : ('data_source', meta['var_name']),
              'var_units'   : ('data_source', meta['var_units']),
              'ts_type_src' : ('data_source', meta['ts_type_src']),
              'time'        : time,
              'latitude'    : lats,
              'longitude'   : lons}
    
    dims = ['data_source', 'time', 'latitude', 'longitude']
    
    data = ColocatedData(data=arr, coords=coords, dims=dims,
                         name=gridded_data.var_name, attrs=meta)
    
    if colocate_time and grid_ts_type != ts_type:
        data = data.resample_time(to_ts_type=ts_type, 
                                  colocate_time=True,
                                  apply_constraints=apply_time_resampling_constraints,
                                  min_num_obs=min_num_obs,
                                  **kwargs)
    return data

def colocate_gridded_ungridded(gridded_data, ungridded_data, ts_type=None, 
                               start=None, stop=None, filter_name=None,
                               regrid_res_deg=None, remove_outliers=True,
                               vert_scheme=None, harmonise_units=True, 
                               var_ref=None, 
                               var_outlier_ranges=None, 
                               var_ref_outlier_ranges=None,
                               update_baseyear_gridded=None,
                               ignore_station_names=None,
                               apply_time_resampling_constraints=None,
                               min_num_obs=None,
                               colocate_time=False,
                               var_keep_outliers=True,
                               var_ref_keep_outliers=False, 
                               **kwargs):
    """Colocate gridded with ungridded data 
    
    Note
    ----
    Uses the variable that is contained in input :class:`GriddedData` object 
    (since these objects only contain a single variable)
    
    Parameters
    ----------
    gridded_data : GriddedData
        gridded data (e.g. model results)
    ungridded_data : UngriddedData
        ungridded data (e.g. observations)
    ts_type : str
        desired temporal resolution of colocated data (must be valid AeroCom
        ts_type str such as daily, monthly, yearly.). The colocation itself is
        done in the highest available resolution and resampling to `ts_type` is
        done afterwards. You may change this behaviour by setting input param
        `resample_first=True` (default is False).
    start : :obj:`str` or :obj:`datetime64` or similar, optional
        start time for colocation, if None, the start time of the input
        :class:`GriddedData` object is used
    stop : :obj:`str` or :obj:`datetime64` or similar, optional
        stop time for colocation, if None, the stop time of the input
        :class:`GriddedData` object is used
    filter_name : str
        string specifying filter used (cf. :class:`pyaerocom.filter.Filter` for
        details). If None, then it is set to 'WORLD-wMOUNTAINS', which 
        corresponds to no filtering (world with mountains). 
        Use WORLD-noMOUNTAINS to exclude mountain sites.
    regrid_res_deg : :obj:`int`, optional
        regrid resolution in degrees. If specified, the input gridded data 
        object will be regridded in lon / lat dimension to the input 
        resolution. (BETA feature)
    remove_outliers : bool
        if True, outliers are removed from model and obs data before colocation, 
        else not.
    vert_scheme : str
        string specifying scheme used to reduce the dimensionality in case 
        input grid data contains vertical dimension. Example schemes are 
        `mean, surface, altitude`, for details see 
        :func:`GriddedData.to_time_series`.
    harmonise_units : bool
        if True, units are attempted to be harmonised (note: raises Exception
        if True and units cannot be harmonised).
    var_ref : :obj:`str`, optional
        variable against which data in :attr:`gridded_data` is supposed to be
        compared. If None, then the same variable is used 
        (i.e. `gridded_data.var_name`).
    var_outlier_ranges : dict, optional
        dictionary specifying outlier ranges for dataset to be analysed
        (e.g. dict(od550aer = [-0.05, 10], ang4487aer=[0,4])). If None, then
        the pyaerocom default outlier ranges are used for the input variable.
        Defaults to None.
    var_ref_outlier_ranges : dict, optional
        like `var_outlier_ranges` but for reference dataset.
    update_baseyear_gridded : int, optional
        optional input that can be set in order to re-define the time dimension
        in the gridded data object to be analysed. E.g., if the data object 
        is a climatology (one year of data) that has set the base year of the
        time dimension to a value other than the specified input start / stop 
        time this may be used to update the time in order to make colocation 
        possible.
    ignore_station_names : str or list, optional
        station name or pattern or list of station names or patterns that should
        be ignored
    apply_time_resampling_constraints : bool, optional
        if True, then time resampling constraints are applied as provided via 
        :attr:`min_num_obs` or if that one is unspecified, as defined in
        :attr:`pyaerocom.const.OBS_MIN_NUM_RESAMPLE`. If None, than 
        :attr:`pyaerocom.const.OBS_APPLY_TIME_RESAMPLE_CONSTRAINTS` is used
        (which defaults to True !!).
    min_num_obs : int or dict, optional
        minimum number of observations for resampling of time
    colocate_time : bool
        if True and if original time resolution of data is higher than desired
        time resolution (`ts_type`), then both datasets are colocated in time 
        *before* resampling to lower resolution. 
    var_keep_outliers : bool
        if True, then no outliers will be removed from dataset to be analysed, 
        even if `remove_outliers` is True. That is because for model evaluation
        often only outliers are supposed to be removed in the observations but
        not in the model.
    var_ref_keep_outliers : bool
        if True, then no outliers will be removed from the reference dataset, 
        even if `remove_outliers` is True.
    **kwargs
        additional keyword args (passed to 
        :func:`UngriddedData.to_station_data_all`)
        
    Returns
    -------
    ColocatedData
        instance of colocated data
        
    Raises
    ------
    VarNotAvailableError
        if grid data variable is not available in ungridded data object
    AttributeError
        if instance of input :class:`UngriddedData` object contains more than
        one dataset
    TimeMatchError
        if gridded data time range does not overlap with input time range
    ColocationError
        if none of the data points in input :class:`UngriddedData` matches 
        the input colocation constraints
    """
    if var_outlier_ranges is None:
        var_outlier_ranges = {}
    if var_ref_outlier_ranges is None:
        var_ref_outlier_ranges = {}
        
    if filter_name is None:
        filter_name = 'WORLD-wMOUNTAINS'
    
    var = gridded_data.var_name
    aerocom_var = gridded_data.var_name_aerocom
    if var_ref is None:
        var_ref = aerocom_var
        
    if remove_outliers:
        low, high, low_ref, high_ref = None, None, None, None    
        if var in var_outlier_ranges:
            low, high = var_outlier_ranges[var]
        if var_ref in var_ref_outlier_ranges:
            low_ref, high_ref = var_ref_outlier_ranges[var_ref]
            
    if not var_ref in ungridded_data.contains_vars:
        raise VarNotAvailableError('Variable {} is not available in ungridded '
                                   'data (which contains {})'
                                   .format(var_ref,
                                           ungridded_data.contains_vars))
    elif len(ungridded_data.contains_datasets) > 1:
        raise AttributeError('Colocation can only be performed with '
                             'ungridded data objects that only contain a '
                             'single dataset. Use method `extract_dataset` of '
                             'UngriddedData object to extract single datasets')
    
    dataset_ref = ungridded_data.contains_datasets[0]
    
    if update_baseyear_gridded is not None:
        # update time dimension in gridded data
        gridded_data.base_year = update_baseyear_gridded
    # get start / stop of gridded data as pandas.Timestamp
    grid_start = to_pandas_timestamp(gridded_data.start)
    grid_stop = to_pandas_timestamp(gridded_data.stop)
    
    grid_ts_type = gridded_data.ts_type
    
    if ts_type is None:
        ts_type = grid_ts_type
    if start is None:
        start = grid_start
    else:
        start = to_pandas_timestamp(start)    
    if stop is None:
        stop = grid_stop
    else:
        stop = to_pandas_timestamp(stop)
    
    if start < grid_start:
        start = grid_start
    if stop > grid_stop:
        stop = grid_stop
    # check overlap
    if stop < grid_start or start > grid_stop:
        raise TimeMatchError('Input time range {}-{} does not '
                             'overlap with data range: {}-{}'
                             .format(start, stop, grid_start, grid_stop))  
    # create instance of Filter class (may, in the future, also include all
    # filter options, e.g. start, stop, variables, only land, only oceans, and
    # may also be linked with other data object, e.g. if data is only supposed
    # to be used if other data object exceeds a certain threshold... but for 
    # now, only region and altitude range)
    regfilter = Filter(name=filter_name)
    
    # apply filter to data
    ungridded_data = regfilter(ungridded_data)

    #crop time
    gridded_data = gridded_data.crop(time_range=(start, stop))
    
    if regrid_res_deg is not None:
        
        lons = gridded_data.longitude.points
        lats = gridded_data.latitude.points
        
        lons_new = np.arange(lons.min(), lons.max(), regrid_res_deg)
        lats_new = np.arange(lats.min(), lats.max(), regrid_res_deg) 
        
        gridded_data = gridded_data.interpolate(latitude=lats_new, 
                                                longitude=lons_new)
    
    ungridded_freq = None # that keeps ungridded data in original resolution
    
    if not colocate_time:
        gridded_data = gridded_data.resample_time(to_ts_type=ts_type)
        ungridded_freq = ts_type # converts ungridded data directly to desired resolution
        
    # ts_type that is used for colocation
    col_ts_type = gridded_data.ts_type
    
    # pandas frequency string that corresponds to col_ts_type
    col_freq = TS_TYPE_TO_PANDAS_FREQ[col_ts_type]
    
    if remove_outliers and not var_ref_keep_outliers:
        ungridded_data.remove_outliers(var_ref, inplace=True,
                                       low=low_ref, 
                                       high=high_ref)
    
    all_stats = ungridded_data.to_station_data_all(
            
            vars_to_convert=var_ref, 
            start=start, 
            stop=stop, 
            freq=ungridded_freq,
            by_station_name=True,
            ignore_index=ignore_station_names,
            apply_constraints=apply_time_resampling_constraints,
            min_num_obs=min_num_obs,
            **kwargs)
    
    obs_stat_data = all_stats['stats']
    ungridded_lons = all_stats['longitude']
    ungridded_lats = all_stats['latitude']
    
    # resampling constraints may have been altered in case input was None, 
    # thus overwrite
    vi = obs_stat_data[0]['var_info'][var_ref]
    if 'apply_constraints' in vi:
        apply_time_resampling_constraints = vi['apply_constraints']
        min_num_obs = vi['min_num_obs']
    
    if len(obs_stat_data) == 0:
        raise VarNotAvailableError('Variable {} is not available in specified '
                                   'time interval ({}-{})'
                                   .format(var_ref, start, stop))
    # make sure the gridded data is in the right dimension
    try:
        gridded_data.check_dimcoords_tseries()
    except DimensionOrderError:
        gridded_data.reorder_dimensions_tseries()
    
    if gridded_data.ndim > 3:
        if vert_scheme is None:
            vert_scheme = 'mean'
        if not vert_scheme in gridded_data.SUPPORTED_VERT_SCHEMES:
            raise ValueError('Vertical scheme {} is not supported'.format(vert_scheme))
            
    grid_stat_data = gridded_data.to_time_series(longitude=ungridded_lons,
                                                 latitude=ungridded_lats,
                                                 vert_scheme=vert_scheme)
    
    # Generate time index of ColocatedData object
    time_idx = pd.DatetimeIndex(start=start, end=stop, freq=col_freq)
    #periods = time_idx.to_period(col_freq)
# =============================================================================
#     if col_freq in PANDAS_RESAMPLE_OFFSETS:
#         offs = np.timedelta64(1, '[{}]'.format(PANDAS_RESAMPLE_OFFSETS[col_freq]))
#         time_idx = time_idx + offs
# =============================================================================
    
    coldata = np.empty((2, len(time_idx), len(obs_stat_data)))
    
    lons = []
    lats = []
    alts = []
    station_names = []
                
    ungridded_unit = None
    ts_type_src_ref = None
    if not harmonise_units:
        gridded_unit = str(gridded_data.units)
    else:
        gridded_unit = None
    
    # loop over all stations and append to colocated data object
    for i, obs_stat in enumerate(obs_stat_data):
        
        if ts_type_src_ref is None:
            ts_type_src_ref = obs_stat['ts_type_src']
        elif obs_stat['ts_type_src'] != ts_type_src_ref:
            spl = ts_type_src_ref.split(';')
            if not obs_stat['ts_type_src'] in spl:
                spl.append(obs_stat['ts_type_src'])
            ts_type_src_ref = ';'.join(spl)
            
        if ungridded_unit is None:
            try:
                ungridded_unit = obs_stat['var_info'][var_ref]['units']
            except KeyError as e: #variable information or unit is not defined
                logger.exception(repr(e))
        try:
            unit = obs_stat['var_info'][var_ref]['units']
        except:
            unit = None
        if not unit == ungridded_unit:
            raise ValueError('Cannot perform colocation. Ungridded data '
                             'object contains different units ({})'.format(var_ref))
        # get observations (Note: the index of the observation time series
        # is already in the specified frequency format, and thus, does not
        # need to be updated, for details (or if errors occur), cf. 
        # UngriddedData.to_station_data, where the conversion happens)
        
        # get model data corresponding to station
        grid_stat = grid_stat_data[i]
        if harmonise_units:
            grid_unit = grid_stat.get_unit(var)
            obs_unit = obs_stat.get_unit(var_ref)
            if not grid_unit == obs_unit:
                grid_stat.convert_unit(var, obs_unit)
            if gridded_unit is None:
                gridded_unit = obs_unit
        
        if remove_outliers and not var_keep_outliers:
            # don't check if harmonise_units is active, because the 
            # remove_outliers method checks units based on AeroCom default 
            # variables, and a variable mapping might be active, i.e. 
            # sometimes models use abs550aer for absorption coefficients 
            # with units [m-1] and not for AAOD (which is the AeroCom default
            # and unitless. Hence, unit check in remove_outliers works only
            # if the variable name (and unit) corresonds to AeroCom default)
            #chk_unit = not harmonise_units 
            grid_stat.remove_outliers(var, low=low, high=high,
                                      check_unit=True)
        
        # get grid and obs timeseries data (that may be sampled in arbitrary
        # time resolution, particularly the obs data)
        grid_ts = grid_stat[var]
        obs_ts = obs_stat[var_ref]
        
        # resample to the colocation frequency
        obs_ts1 = obs_ts.resample(col_freq).mean()
        grid_ts1 = grid_ts.resample(col_freq).mean()
        
        # fill up missing time stamps
        _df = pd.concat([obs_ts1, grid_ts1], axis=1, keys=['o', 'm'])
        
        # assign the unified timeseries data to the colocated data array
        coldata[0, :, i] = _df['o'].values
        coldata[1, :, i] = _df['m'].values
        
        lons.append(obs_stat.longitude)
        lats.append(obs_stat.latitude)
        alts.append(obs_stat.altitude)
        station_names.append(obs_stat.station_name)
    
    try:
        revision = ungridded_data.data_revision[dataset_ref]
    except: 
        try:
            revision = ungridded_data._get_data_revision_helper(dataset_ref)
        except MetaDataError:
            revision = 'MULTIPLE'
        except:
            revision = 'n/a'
            
    files = [os.path.basename(x) for x in gridded_data.from_files]
    
    meta = {'data_source'       :   [dataset_ref,
                                     gridded_data.name],
            'var_name'          :   [var_ref, var],
            'ts_type'           :   col_ts_type,
            'filter_name'       :   filter_name,
            'ts_type_src'       :   [ts_type_src_ref, grid_ts_type],
            'start_str'         :   to_datestring_YYYYMMDD(start),
            'stop_str'          :   to_datestring_YYYYMMDD(stop),
            'var_units'         :   [ungridded_unit,
                                     gridded_unit],
            'vert_scheme'       :   vert_scheme,
            'data_level'        :   3,
            'revision_ref'      :   revision,
            'from_files'        :   files,
            'from_files_ref'    :   None,
            'stations_ignored'  :   ignore_station_names,
            'colocate_time'     :   colocate_time,
            'apply_constraints' :   apply_time_resampling_constraints,
            'min_num_obs'       :   min_num_obs,
            'outliers_removed'  :   remove_outliers}

    
    meta.update(regfilter.to_dict())
    
    # create coordinates of DataArray
    coords = {'data_source' : meta['data_source'],
              'var_name'    : ('data_source', meta['var_name']),
              'var_units'   : ('data_source', meta['var_units']),
              'ts_type_src' : ('data_source', meta['ts_type_src']),
              'time'        : time_idx,
              'station_name': station_names,
              'latitude'    : ('station_name', lats),
              'longitude'   : ('station_name', lons),
              'altitude'    : ('station_name', alts)
              }
    
    dims = ['data_source', 'time', 'station_name']
    data = ColocatedData(data=coldata, coords=coords, dims=dims, name=var,
                         attrs=meta)
    
    if colocate_time and grid_ts_type != ts_type:
        data = data.resample_time(to_ts_type=ts_type, 
                                  colocate_time=True,
                                  apply_constraints=apply_time_resampling_constraints,
                                  min_num_obs=min_num_obs,
                                  **kwargs)
    return data

if __name__=='__main__':
    import pyaerocom as pya
    import matplotlib.pyplot as plt
    plt.close('all')
    
    obsdata = pya.io.ReadGridded('MODIS6.aqua').read_var('od550aer', 
                                start=2010)
    
    
    modeldata = pya.io.ReadGridded('ECMWF_CAMS_REAN').read_var('od550aer')
    coldata = pya.colocation.colocate_gridded_gridded(modeldata, obsdata, 
                                                      ts_type='monthly',
                                                      regrid_res_deg=5,
                                                      remove_outliers=True,
                                                      colocate_time=False)
    
    stats = coldata.calc_statistics()
    
    coldata_alt = pya.colocation.colocate_gridded_gridded(modeldata, obsdata, 
                                                      ts_type='monthly',
                                                      regrid_res_deg=5,
                                                      remove_outliers=True,
                                                      colocate_time=True)
    
    stats_alt = coldata_alt.calc_statistics()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains plot routines for Aerocom data. So far, this includes a
low level, very flexible plot method
"""
from matplotlib.pyplot import figure, get_cmap
from pandas import to_datetime
import numpy as np
from matplotlib.colors import BoundaryNorm, LogNorm, Normalize

import cartopy.crs as ccrs
from cartopy.mpl.geoaxes import GeoAxes
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
from numpy import meshgrid, linspace, ceil

from pyaerocom import logger
from pyaerocom.plot.config import COLOR_THEME, ColorTheme, MAP_AXES_ASPECT
from pyaerocom.plot.helpers import (custom_mpl, 
                                    calc_pseudolog_cmaplevels,
                                    projection_from_str,
                                    calc_figsize)
from pyaerocom.mathutils import exponent
from pyaerocom.region import Region

MPL_PARAMS = custom_mpl()

def get_cmap_maps_aerocom(color_theme=None, vmin=None, vmax=None):
    """Get colormap using pyAeroCom color scheme
    
    Parameters
    ----------
    color_theme : :obj:`ColorTheme`, optional
        instance of pyaerocom color theme. If None, the default schemes is used
    vmin : :obj:`float`, optional
        lower end of value range
    vmax : :obj:`float`, optional
        upper end of value range
    
    Returns
    -------
    colormap
    """
    if color_theme is None:
        color_theme=COLOR_THEME
    if vmin is not None and vmax is not None and vmin < 0 and vmax > 0:
        cmap = get_cmap(color_theme.cmap_map_div)
        if color_theme.cmap_map_div_shifted:
            try:
                from geonum.helpers import shifted_color_map
                cmap = shifted_color_map(vmin, vmax, cmap)
            except:
                logger.warning('cannot shift colormap, need geonum installation')
        return cmap
    return get_cmap(color_theme.cmap_map)
    

def set_map_ticks(ax, xticks=None, yticks=None):
    """Set or update ticks in instance of GeoAxes object (cartopy)
    
    Parameters
    ----------
    ax : cartopy.GeoAxes
        map axes instance
    xticks : iterable, optional
        ticks of x-axis (longitudes)
    yticks : iterable, optional
        ticks of y-axis (latitudes)
        
    Returns
    -------
    cartopy.GeoAxes
        modified axes instance
    """
    lonleft, lonright = ax.get_xlim()
    digits = 2 - exponent(lonleft)
    digits = 0 if digits < 0 else digits
    tick_format = '.%df' %digits
    if not xticks:    
        num_lonticks = 7 if lonleft == -lonright else 6    
        xticks = linspace(lonleft, lonright, num_lonticks)
    if not yticks:
        latleft, latright = ax.get_ylim()
        num_latticks = 7 if latleft == - latright else 6
        yticks = linspace(latleft, latright, num_latticks)
    ax.set_xticks(xticks, crs=ccrs.PlateCarree())
    ax.set_yticks(yticks, crs=ccrs.PlateCarree())    

    lon_formatter = LongitudeFormatter(number_format=tick_format,
                                       degree_symbol='',
                                       dateline_direction_label=True)
    lat_formatter = LatitudeFormatter(number_format=tick_format,
                                      degree_symbol='')
    
    ax.xaxis.set_major_formatter(lon_formatter)
    ax.yaxis.set_major_formatter(lat_formatter)
    
    return ax
    
def init_map(xlim=(-180, 180), ylim=(-90, 90), figh=8, fix_aspect=False,
             xticks=None, yticks=None, color_theme=COLOR_THEME, 
             projection=None, title=None, gridlines=False, 
             fig=None, ax=None, draw_coastlines=True, contains_cbar=False):
    """Initalise a map plot
    
    Parameters
    ----------
    xlim : tuple
        2-element tuple specifying plotted longitude range
    ylim : tuple
        2-element tuple specifying plotted latitude range
    figh : int
        height of figure in inches
    fix_aspect : :obj:`float`, optional
        if True, the aspect of the GeoAxes instance is kept fix using the
        default aspect ``MAP_AXES_ASPECT`` defined in 
        :mod:`pyaerocom.plot.config`
    xticks : iterable, optional
        ticks of x-axis (longitudes)
    yticks : iterable, optional
        ticks of y-axis (latitudes)
    projection 
        projection instance from cartopy.crs module (e.g. PlateCaree). May also
        be string
    title : :obj:`str`, optional
        title that is supposed to be inserted
    fig : :obj:`Figure`, optional
        instance of matplotlib Figure class. If specified, the former to 
        input args (``figh`` and ``fix_aspect``) are ignored. Note that the 
        Figure is wiped clean before plotting, so any plotted content will be 
        lost
    
    Returns
    -------
    ax : cartopy.mpl.geoaxes.GeoAxes
        axes instance
    """
    
    if projection is None:
        projection = ccrs.PlateCarree()
    elif isinstance(projection, str):
        projection = projection_from_str(projection)
    elif not isinstance(projection, ccrs.Projection):
        raise ValueError("Input for projection needs to be instance of "
                         "cartopy.crs.Projection")
    
    if not isinstance(ax, GeoAxes):
        if fig is None:
            if not fix_aspect:
                figsize = calc_figsize(xlim, ylim)
                #figw = figh*2
            else:
                figw = figh*fix_aspect
                figsize = (figw, figh)
                    
            fig = figure(figsize=figsize)
        else:
            fig.clf()
        if contains_cbar:
            ax = fig.add_axes([0.1, 0.12, 0.75, 0.8], projection=projection)
        else:
            ax = fig.add_axes([0.1, 0.12, 0.85, 0.8], projection=projection)
    
    if fix_aspect:
        ax.set_aspect(MAP_AXES_ASPECT)
    
    if not isinstance(color_theme, ColorTheme):
        if isinstance(color_theme, str):
            color_theme = ColorTheme(color_theme)  
        else:
            color_theme = COLOR_THEME
            
    if draw_coastlines:
        ax.coastlines(color=COLOR_THEME.color_coastline)
    
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    
    ax = set_map_ticks(ax, xticks, yticks)
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    
    if title is not None:
        ax.set_title(title)
    if gridlines:
        ax.gridlines()
    
    return ax

def plot_griddeddata_on_map(data, lons=None, lats=None, var_name=None, 
                            unit=None, xlim=(-180, 180), ylim=(-90, 90), 
                            vmin=None, vmax=None, add_zero=False, c_under=None, 
                            c_over=None, log_scale=True, discrete_norm=True, 
                            cbar_levels=None, cbar_ticks=None, add_cbar=True,
                            cmap=None, cbar_ticks_sci=False, 
                            color_theme=COLOR_THEME, **kwargs):
    """Make a plot of gridded data onto a map
    
    Note
    ----
    This is a lowlevel plotting method
    
    Parameters
    ----------
    data : ndarray
        2D data array
    lons : ndarray
        longitudes of data
    lats : ndarray
        latitudes of data
    var_name : :obj:`str`, optional
        name of variable that is plotted
    xlim : tuple
        2-element tuple specifying plotted longitude range
    ylim : tuple
        2-element tuple specifying plotted latitude range
    vmin : :obj:`float`, optional
        lower value of colorbar range
    vmax : :obj:`float`, optional 
        upper value of colorbar range
    add_zero : bool
        if True and vmin is not 0, then, the colorbar is extended down to 0. 
        This may be used, e.g. for logarithmic scales that should include 0.
    c_under : :obj:`float`, optional 
        colour of data values smaller than ``vmin``
    c_over : :obj:`float`, optional 
        colour of data values exceeding ``vmax``
    log_scale : bool
        if True, the value to color mapping is done in a pseudo log scale 
        (see :func:`get_cmap_levels_auto` for implementation)
    discrete_norm : bool
        if True, color mapping will be subdivided into discrete intervals
    cbar_levels : iterable, optional
        discrete colorbar levels. Will be computed automatically, if None 
        (and applicable)
    cbar_ticks : iterable, optional
        ticks of colorbar levels. Will be computed automatically, if None 
        (and applicable)
    
    Returns
    -------
    fig
        matplotlib figure instance containing plot result. Use 
        ``fig.axes[0]`` to access the map axes instance (e.g. to modify the 
        title or lon / lat range, etc.)
    """
    kwargs['contains_cbar'] = True
    ax = init_map(xlim, ylim, color_theme=color_theme, **kwargs)
    fig = ax.figure
    from pyaerocom.griddeddata import GriddedData
    if isinstance(data, GriddedData):
        if not data.has_latlon_dims:
            from pyaerocom.exceptions import DataDimensionError
            raise DataDimensionError('Input data needs to have latitude and '
                                     'longitude dimension')
        if not data.ndim == 2:
            if not data.ndim == 3 or not 'time' in data.dimcoord_names:
                raise DataDimensionError('Input data needs to be 2 dimensional '
                                         'or 3D with time being the 3rd '
                                         'dimension')
            data.reorder_dimensions_tseries()
            
            data = data[0]
        
        lons = data.longitude.points
        lats = data.latitude.points
        data = data.grid.data
    elif not isinstance(data, np.ndarray) or not data.ndim == 2:
        raise IOError("Need 2D numpy array")
    elif not isinstance(lats, np.ndarray) or not isinstance(lons, np.ndarray):
        raise ValueError('Missing lats or lons input')
    if isinstance(data, np.ma.MaskedArray):
        sh = data.shape
        if data.mask.sum() == sh[0] * sh[1]:
            raise ValueError('All datapoints in input data (masked array) are '
                             'invalid')
    _loc = ax.bbox._bbox
    try:
        ax_cbar = fig.add_axes([_loc.x1 + .02,
                                _loc.y0, .02, _loc.y1 - _loc.y0])
    except Exception as e:
        ax_cbar = fig.add_axes([0.91, 0.12, .02, .8])
        print(repr(e))
    X, Y = meshgrid(lons, lats)
    dmin = np.nanmin(data)
    dmax = np.nanmax(data)
    
    if any([np.isnan(x) for x in [dmin, dmax]]):
        raise ValueError('Cannot plot map of data: all values are NaN')
    elif dmin == dmax:
        raise ValueError('Minimum value in data equals maximum value: '
                         '{}'.format(dmin))
    if vmin is None:
        vmin = dmin
    else:
        if vmin < 0 and log_scale:
            log_scale=False
    if vmax is None:
        vmax = dmax
    
    bounds = None
    if cbar_levels: #user provided levels of colorbar explicitely
        if vmin is not None or vmax is not None:
            raise ValueError('Please provide either vmin/vmax OR cbar_levels')
        bounds = list(cbar_levels)
        low, high = bounds[0], bounds[-1]
        if add_zero and low > 0:
            bounds.insert(0, 0) # insert zero bound
        if cmap is None or isinstance(cmap, str):
            cmap = get_cmap_maps_aerocom(color_theme, low, high)
        norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)
    else:
        if log_scale: # no negative values allowed
            if vmin < 0:
                vmin = data[data>0].min()
                if c_under is None: #special case, set c_under to indicate that there is values below 0
                    c_under = 'r'
            if cmap is None or isinstance(cmap, str):
                cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
            if discrete_norm:
                #to compute upper range of colour range, round up vmax
                exp = float(exponent(vmax) - 1)
                vmax_colors = ceil(vmax / 10**exp)*10**exp
                bounds = calc_pseudolog_cmaplevels(vmin=vmin, vmax=vmax_colors,
                                                   add_zero=add_zero)
                norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)
        
            else:
                norm = LogNorm(vmin=vmin, vmax=vmax, clip=True)
        else: 
            if cmap is None or isinstance(cmap, str):
                cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
            norm = Normalize(vmin=vmin, vmax=vmax)
    cbar_extend = "neither"
    if c_under is not None:
        cmap.set_under(c_under)
        cbar_extend = "min"
        if bounds is not None:
            bounds.insert(0, bounds[0] - bounds[1])
    if c_over is not None:
        cmap.set_over(c_over)
        if bounds is not None:
            bounds.append(bounds[-1] + bounds[-2])
        if cbar_extend == "min":
            cbar_extend = "both"
        else:
            cbar_extend = "max"
      
    disp = ax.pcolormesh(X, Y, data, cmap=cmap, norm=norm)
# =============================================================================
#     fmt = None
#     if bounds is not None:
#         print(bounds)
#         min_mag = -exponent(bounds[1])
#         min_mag = 0 if min_mag < 0 else min_mag
#         print(min_mag)
#         #fmt = "%." + str(min_mag) + "f"
# =============================================================================
    if add_cbar:
        cbar = fig.colorbar(disp, cmap=cmap, norm=norm, boundaries=bounds, 
                            extend=cbar_extend, cax=ax_cbar)
        
        if var_name is not None:
            var_str = var_name# + VARS.unit_str
            if unit is not None:
                if not str(unit) in ['1', 'no_unit']:
                    var_str += ' [{}]'.format(unit)
            
            cbar.set_label(var_str)
        
        if cbar_ticks:
            cbar.set_ticks(cbar_ticks)
        if cbar_ticks_sci:
            lbls = []
            for lbl in cbar.ax.get_yticklabels():
                tstr = lbl.get_text()
                if bool(tstr):
                    lbls.append('{:.1e}'.format(float(tstr)))
                else:
                    lbls.append('')
            cbar.ax.set_yticklabels(lbls)
            
    return fig

def plot_map_aerocom(data, region=None, fig=None, **kwargs):
    """High level map plotting function for Aerocom default plotting
    
    Note
    ----
    This function does not iterate over a cube in time, but uses the
    first available time index in the data.
    
    Parameters
    ----------
    data : :obj:`GriddedData`
        input data from one timestamp (if data contains more than one time 
        stamp, the first index is used)
        
    TODO 
        finish docstring
    
        
    """
    #kwargs["fix_aspect"] = 1.6
    from pyaerocom import GriddedData
    if not isinstance(data, GriddedData):
        raise TypeError("This plotting method needs an instance of pyaerocom "
                         "GriddedData on input, got: %s" %type(data))
    if region:
        if isinstance(region, str):
            region = Region(region)
        if not isinstance(region, Region):
            raise TypeError("Invalid input for region, need None, str or Region")    
        data = data.crop(region=region)
    if not data.suppl_info["region"]:
        data = data.crop(region="WORLD")
    region = data.suppl_info["region"]
    s = data.plot_settings
    fig = plot_griddeddata_on_map(data, 
                                  xlim=region.lon_range_plot, 
                                  ylim=region.lat_range_plot,
                                  vmin=s.map_vmin, vmax=s.map_vmax, 
                                  c_over=s.map_c_over, c_under=s.map_c_under,
                                  cbar_levels=s.map_cbar_levels, 
                                  xticks = region.lon_ticks,
                                  yticks = region.lat_ticks,
                                  cbar_ticks=s.map_cbar_ticks, **kwargs)
    ax = fig.axes[0]
    
    # annotate model in lower left corner
    lonr, latr = region.lon_range_plot, region.lat_range_plot
    ax.annotate(data.name,
                xy=(lonr[0] + (lonr[1] - lonr[0])*.03,
                    latr[0] + (latr[1] - latr[0])*.03), 
                xycoords='data', 
                horizontalalignment='left', 
                color='black', 
                fontsize=MPL_PARAMS['axes.titlesize'] + 2,
                bbox=dict(boxstyle='square', 
                          facecolor='white', 
                          edgecolor='none',
                          alpha=0.7))
    ax.annotate('source: AEROCOM', 
                xy=(0.97, 0.03), 
                xycoords='figure fraction',
                horizontalalignment='right', 
                fontsize=MPL_PARAMS['xtick.labelsize'], 
                bbox=dict(boxstyle='square', 
                          facecolor='none',
                          edgecolor='black'))
    ax.set_title("{} {} mean {:.3f}".format(data.var_name.upper(), 
                 to_datetime(data.start).strftime("%Y%m%d"),
                 data.area_weighted_mean()))
    return fig
    
def plot_map(data, *args, **kwargs):
    """Map plot of grid data
    
    Note
    ----
    Deprecated name of method. Please use :func:`plot_griddeddata_on_map` in 
    the future.
    
    Parameters
    ----------
    data 
        data (2D numpy array or instance of GriddedData class. The latter is
        deprecated, but will continue to work)    
    *args, **kwargs
        See :func:`plot_griddeddata_on_map`
    
    Returns
    -------
    See :func:`plot_griddeddata_on_map`
    """
    from pyaerocom import print_log, GriddedData
    print_log.warning(DeprecationWarning('Method name plot_map is deprecated. '
                                         'Please use plot_griddeddata_on_map'))
    if isinstance(data, GriddedData):
        if 'time' in data and len(data['time'].points) > 1:
            logger.warning("Input data contains more than one time stamp, using "
                           "first time stamp")
            data = data[0]
        if not all([x in data for x in ('longitude', 'latitude')]):
            raise AttributeError('GriddedData does not contain either longitude '
                                 'or latitude coordinates')
        return plot_griddeddata_on_map(data.grid.data,
                                       data.longitude.points,
                                       data.latitude.points, 
                                       *args, **kwargs)
    return plot_griddeddata_on_map(data, *args, **kwargs)
   
if __name__ == "__main__":
    from matplotlib.pyplot import close
    import pyaerocom as pya
    close("all")
    
    d = pya.io.ReadGridded('ECMWF-IFS-CY42R1-CAMS-RA-CTRL_AP3-CTRL2016-PD').read_var('ang4487aer', start=2010)
    d.quickplot_map(vmin=-1, vmax=4)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:32:52 2018

@author: jonasg
"""
import ipywidgets as ipw
from IPython.display import display

def ProgressBarLabelled(num, label=''):
    try:
        f = ipw.IntProgress(0, max=num)
        l = ipw.Label(label)
        display(ipw.HBox([l, f]))
        return f
    except:
        pass



################################################################
# read_earlinet.py
#
# read Earlinet lidar profile data
#
# this file is part of the pyaerocom package
#
# Copyright (C) 2017 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jonasg@met.no
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA

import os, fnmatch, re
from collections import OrderedDict as od
import numpy as np
import xarray
from pyaerocom import const
from pyaerocom.units_helpers import unit_conversion_fac
from pyaerocom.exceptions import DataUnitError
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom import StationData, VerticalProfile, Variable
from pyaerocom import UngriddedData
        
# TODO: Check station names -> they are NOT UNIQUE (e.g. Potenza...) -> maybe
# use station_id instead... would require more flexible iterator in 
# UngriddedData
class ReadEarlinet(ReadUngriddedBase):
    """Interface for reading of EARLINET data"""
    #: Mask for identifying datafiles 
    _FILEMASK = '*.*'
    
    #: version log of this class (for caching)
    __version__ = "0.14_" + ReadUngriddedBase.__baseversion__
    
    #: Name of dataset (OBS_ID)
    DATA_ID = const.EARLINET_NAME
    
    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.EARLINET_NAME]
    
    #: default variables for read method
    DEFAULT_VARS = ['ec532aer']
    
    #: all data values that exceed this number will be set to NaN on read. This
    #: is because iris, xarray, etc. assign a FILL VALUE of the order of e36 
    #: to missing data in the netcdf files
    _MAX_VAL_NAN = 1e6
    
    #: variable name of altitude in files
    ALTITUDE_ID = 'Altitude'
    
    #: temporal resolution
    TS_TYPE = 'n/d'
    
    #: dictionary specifying the file search patterns for each variable
    VAR_PATTERNS_FILE = {'ec532aer'     : '*.e532', 
                         'ec355aer'     : '*.e355',
                         'bscatc532aer' : '*.b532',
                         'bscatc355aer' : '*.b355',
                         'bscatc1064aer': '*.b1064',
                         'zdust'        : '*.e*'}
    
    
    #: dictionary specifying the file column names (values) for each Aerocom 
    #: variable (keys)
    VAR_NAMES_FILE = {'ec532aer'        : 'Extinction', 
                      'ec355aer'        : 'Extinction', 
                      'ec1064aer'       : 'Extinction',
                      'bscatc532aer'    : 'Backscatter',
                      'bscatc355aer'    : 'Backscatter',
                      'bscatc1064aer'   : 'Backscatter',
                      'zdust'           : 'DustLayerHeight'}
        
    #: metadata names that are supposed to be imported
    META_NAMES_FILE = od(location           = 'Location',
                         start_date         = 'StartDate',
                         start_utc          = 'StartTime_UT',
                         stop_utc           = 'StopTime_UT',
                         longitude          = 'Longitude_degrees_east',
                         latitude           = 'Latitude_degrees_north',
                         wavelength_emis    = 'EmissionWavelength_nm',
                         wavelength_det     = 'DetectionWavelength_nm',
                         res_raw_m          = 'ResolutionRaw_meter',
                         zenith_ang_deg     = 'ZenithAngle_degrees',
                         instrument_name    = 'System',
                         comments           = 'Comments',
                         shots_avg          = 'ShotsAveraged',
                         detection_mode     = 'DetectionMode',
                         res_eval           = 'ResolutionEvaluated',
                         input_params       = 'InputParameters',
                         altitude           = 'Altitude_meter_asl',
                         eval_method        = 'EvaluationMethod')
    
    #: metadata keys that are needed for reading (must be values in 
    #: :attr:`META_NAMES_FILE`)
    META_NEEDED = ['Location', 'StartDate', 'StartTime_UT', 'StopTime_UT', 
                   'Longitude_degrees_east', 'Latitude_degrees_north',
                   'Altitude_meter_asl']
    
    #: Metadata keys from :attr:`META_NAMES_FILE` that are additional to 
    #: standard keys defined in :class:`StationMetaData` and that are supposed 
    #: to be inserted into :class:`UngriddedData` object created in :func:`read`
    KEEP_ADD_META = ['location', 'wavelength_emis', 'wavelength_det', 
                     'res_raw_m', 'zenith_ang_deg', 'comments', 'shots_avg', 
                     'detection_mode', 'res_eval', 'input_params', 
                     'eval_method']
    
    #: Attribute access names for unit reading of variable data
    VAR_UNIT_NAMES = od(Extinction    = ['ExtinctionUnits', 'units'],
                        Backscatter   = ['BackscatterUnits', 'units'],
                        Altitude      = 'units')
    #: Variable names of uncertainty data
    ERR_VARNAMES = od(  ec532aer = 'ErrorExtinction',
                        ec355aer = 'ErrorExtinction')
    
    #: If true, the uncertainties are also read (where available, cf. ERR_VARNAMES)
    READ_ERR = True
    
    PROVIDES_VARIABLES = list(VAR_PATTERNS_FILE)
    
# =============================================================================
#     EXCLUDE_FILES = ['cirrus.txt', 
#                      'etna.txt', 
#                      'forest_fires.txt', 
#                      'saharan_dust.txt']
# =============================================================================
    EXCLUDE_CASES = ['cirrus.txt']
    
    def __init__(self, dataset_to_read=None):
        # initiate base class
        super(ReadEarlinet, self).__init__(dataset_to_read)
        # make sure everything is properly set up
        if not all([x in self.VAR_PATTERNS_FILE for x in self.PROVIDES_VARIABLES]):
            raise AttributeError("Please specify file search masks in "
                                 "header dict VAR_PATTERNS_FILE for each "
                                 "variable defined in PROVIDES_VARIABLES")
        elif not all([x in self.VAR_NAMES_FILE for x in self.PROVIDES_VARIABLES]):
            raise AttributeError("Please specify file search masks in "
                                 "header dict VAR_NAMES_FILE for each "
                                 "variable defined in PROVIDES_VARIABLES")
        #: private dictionary containing loaded Variable instances, 
        self._var_info = {}
        
        #: files that are supposed to be excluded from reading
        self.exclude_files = []
        
        #: files that were actually excluded from reading
        self.excluded_files = []
     
    
    def read_file(self, filename, vars_to_retrieve=None, read_err=None,
                  remove_outliers=True):
        """Read EARLINET file and return it as instance of :class:`StationData`
        
        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        read_err : bool
            if True, uncertainty data is also read (where available).
        remove_outliers : bool
            if True, outliers are removed for each variable using the
            `minimum` and `maximum` attributes for that variable (accessed 
            via pyaerocom.const.VARS[var_name]).
        
        Returns
        -------
        StationData 
            dict-like object containing results
        """
        if read_err is None: #use default setting
            read_err = self.READ_ERR
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        _vars = []
        for var in vars_to_retrieve:
            if var in self.VAR_PATTERNS_FILE: #make sure to only read what is supported by this file
                if fnmatch.fnmatch(filename, self.VAR_PATTERNS_FILE[var]):
                    _vars.append(var)
            else: # will be computed
                _vars.append(var)
                
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(_vars)
        
        if len(vars_to_compute) > 0:
            raise NotImplementedError("This feature has not yet implemented, as "
                                      "it was not required so far. The "
                                      "implementation requires handling of "
                                      "profile data as well")
        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out['station_id'] = filename.split('/')[-2]
        data_out['data_id'] = self.DATA_ID
        data_out['ts_type'] = self.TS_TYPE
           
        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            if not var in self._var_info:
                self._var_info[var] = Variable(var)
        var_info = self._var_info
            
        
        # Iterate over the lines of the file
        self.logger.debug("Reading file {}".format(filename))
        
        data_in = xarray.open_dataset(filename)
        
        for k, v in self.META_NAMES_FILE.items():
            if v in self.META_NEEDED:
                _meta = data_in.attrs[v]
            else:
                try:
                    _meta = data_in.attrs[v]
                except:
                    _meta = None
            data_out[k] = _meta
                
        
        data_out['station_name'] = re.split('\s|,', data_out['location'])[0].strip()

        str_dummy = str(data_in.StartDate)
        year, month, day = str_dummy[0:4], str_dummy[4:6], str_dummy[6:8]

        str_dummy = str(data_in.StartTime_UT).zfill(6)
        hours, minutes, seconds = str_dummy[0:2], str_dummy[2:4], str_dummy[4:6]
        
        datestring = '-'.join([year, month, day])
        startstring = 'T'.join([datestring, ':'.join([hours, minutes, seconds])])
         
        dtime = np.datetime64(startstring)
        
        str_dummy = str(data_in.StopTime_UT).zfill(6)
        hours, minutes, seconds = str_dummy[0:2], str_dummy[2:4], str_dummy[4:6]
        
        stopstring = 'T'.join([datestring, ':'.join([hours, minutes, seconds])])
        
        stop = np.datetime64(stopstring)
        
        # in case measurement goes over midnight into a new day
        if stop < dtime:
            stop = stop + np.timedelta64(1, '[D]')
            
        data_out['dtime'] = [dtime]
        data_out['stopdtime'] = [stop]
        data_out['has_zdust'] = False
        contains_vars = []
        
        for var in vars_to_read:
            data_out['var_info'][var] = od()
            err_read = False
            unit_ok = False
            outliers_removed = False
            
            netcdf_var_name = self.VAR_NAMES_FILE[var]
            # check if the desired variable is in the file
            if netcdf_var_name not in data_in.variables:
                self.logger.warning("Variable {} not found in file {}".format(var, filename))
                continue
            
            info = var_info[var]
            # xarray.DataArray
            arr = data_in.variables[netcdf_var_name]
            # the actual data as numpy array (or float if 0-D data, e.g. zdust)
            val = np.float64(arr)
            
            # CONVERT UNIT
            unit = None
            
            unames = self.VAR_UNIT_NAMES[netcdf_var_name]
            for u in unames:
                if u in arr.attrs:
                    unit = arr.attrs[u]
            if unit is None:
                raise DataUnitError('Unit of {} could not be accessed in file '
                                    '{}'.format(var, filename))
            unit_fac = None
            try:
                to_unit = self._var_info[var].units
                unit_fac = unit_conversion_fac(unit, to_unit)
                val *= unit_fac
                unit = to_unit
                unit_ok = True
            except Exception as e:
                const.print_log.warning('Failed to convert unit of {} in file '
                                        '{} (Earlinet): Error: {}'
                                        .format(var, filename, repr(e)))
                
            # import errors if applicable
            err = np.nan
            if read_err and var in self.ERR_VARNAMES:
                err_name = self.ERR_VARNAMES[var]
                if err_name in data_in.variables:
                    err = np.float64(data_in.variables[err_name])
                    if unit_ok:
                        err *= unit_fac
                    err_read = True
                    
            # 1D variable
            if var == 'zdust':
                if not val.ndim == 0:
                    raise ValueError('Fatal: dust layer height data must be '
                                     'single value')
        
                if unit_ok and info.minimum < val < info.maximum:
                    const.print_log.warning('zdust value {} out of range, '
                                            'setting to NaN'.format(val))
                    val = np.nan
                    err = np.nan
                if err > self._MAX_VAL_NAN:
                    err = np.nan
                        
                        
                if np.isnan(val):
                    self.logger.warning("Invalid value of variable zdust "
                                        "in file {}. Skipping...!".format(filename))
                    continue
               
                data_out['has_zdust'] = True
                data_out[var] = val
                

            #elif var.startswith('ec'):
            else:
                if not val.ndim == 1:
                    raise ValueError('Extinction data must be one dimensional')
                elif len(val) == 0:
                    continue # no data
                # Remove NaN equivalent values
                val[val>self._MAX_VAL_NAN] = np.nan
                    
                wvlg = var_info[var].wavelength_nm
                wvlg_str = self.META_NAMES_FILE['wavelength_emis']
            
                if not wvlg == data_in.attrs[wvlg_str]:
                    self.logger.info('No wavelength match')
                    continue
                
                alt_id = self.ALTITUDE_ID
                alt_data = data_in.variables[alt_id]
                
                alt_vals = np.float64(alt_data)
                alt_unit = alt_data.attrs[self.VAR_UNIT_NAMES[alt_id]]
                to_alt_unit = const.VARS['alt'].units
                if not alt_unit == to_alt_unit:
                    try:
                        alt_unit_fac = unit_conversion_fac(alt_unit, 
                                                           to_alt_unit)
                        alt_vals *= alt_unit_fac
                        alt_unit = to_alt_unit
                    except Exception as e:
                        self.logger.warning('Failed to convert unit: {}'.format(repr(e)))
                
                # remove outliers from data, if applicable
                if remove_outliers and unit_ok:
                    # REMOVE OUTLIERS
                    outlier_mask = np.logical_or(val < info.minimum, 
                                                 val > info.maximum)
                    val[outlier_mask] = np.nan
                    
                    if err_read:
                        err[outlier_mask] = np.nan
                    outliers_removed = True
                # remove outliers from errors if applicable
                if err_read:
                    err[err > self._MAX_VAL_NAN] = np.nan
                        
                # create instance of ProfileData
                profile = VerticalProfile(data=val,
                                          altitude=alt_vals,
                                          dtime=dtime,
                                          var_name=var,
                                          data_err=err,
                                          var_unit=unit,
                                          altitude_unit=alt_unit)   
                # Write everything into profile
                data_out[var] = profile
                
            
            contains_vars.append(var)
            data_out['var_info'][var].update(unit_ok=unit_ok, 
                                             err_read=err_read,
                                             outliers_removed=outliers_removed)
            
        data_out['contains_vars'] = contains_vars
        return (data_out)
    
    def read(self, vars_to_retrieve=None, files=None, first_file=None, 
             last_file=None, read_err=None, remove_outliers=True,
             file_pattern=None):
        """Method that reads list of files as instance of :class:`UngriddedData`
        
        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None, 
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file 
            in the list is used
        read_err : bool
            if True, uncertainty data is also read (where available). If 
            unspecified (None), then the default is used (cf. :attr:`READ_ERR`)
         file_pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)
            
        Returns
        -------
        UngriddedData
            data object
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        if read_err is None:
            read_err = self.READ_ERR
            
        if files is None:
            if len(self.files) == 0:
                self.get_file_list(vars_to_retrieve, file_pattern=file_pattern)
            files = self.files
    
        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)
        
        files = files[first_file:last_file]
        
        self.read_failed = []
        
        data_obj = UngriddedData()
        col_idx = data_obj.index
        meta_key = -1.0
        idx = 0
        
        #assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx
        
        #last_station_id = ''
        num_files = len(files)
        
        disp_each = int(num_files*0.1)
        if disp_each < 1:
            disp_each = 1
        
        VAR_IDX = -1
        for i, _file in enumerate(files):
            if i%disp_each == 0:
                print("Reading file {} of {} ({})".format(i+1, 
                                 num_files, type(self).__name__))
            try:
                stat = self.read_file(_file, 
                                      vars_to_retrieve=vars_to_retrieve,
                                      read_err=read_err, 
                                      remove_outliers=remove_outliers)
                if not any([var in stat.contains_vars for var in 
                            vars_to_retrieve]):
                    self.logger.info("Station {} contains none of the desired "
                                     "variables. Skipping station..."
                                     .format(stat.station_name))
                    continue
                #if last_station_id != station_id:
                meta_key += 1
                # Fill the metatdata dict
                # the location in the data set is time step dependant!
                # use the lat location here since we have to choose one location
                # in the time series plot
                metadata[meta_key] = od()
                metadata[meta_key].update(stat.get_meta())
                for add_meta in self.KEEP_ADD_META:
                    if add_meta in stat:
                        metadata[meta_key][add_meta] = stat[add_meta]
                #metadata[meta_key]['station_id'] = station_id
                #metadata[meta_key]['data_id'] = self.DATA_ID
                metadata[meta_key]['data_revision'] = self.data_revision
                metadata[meta_key]['variables'] = []
                metadata[meta_key]['var_info'] = od()
                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = od()
                    #last_station_id = station_id
                
                # Is floating point single value
                time = stat.dtime[0]
                for var in stat.contains_vars:
                    if not var in data_obj.var_idx:
                        VAR_IDX +=1
                        data_obj.var_idx[var] = VAR_IDX
                    
                    var_idx = data_obj.var_idx[var]
                        
                    val = stat[var]
                    metadata[meta_key]['var_info'][var] = vi = od()
                    if isinstance(val, VerticalProfile):
                        altitude = val.altitude
                        data = val.data
                        add = len(data)
                        err = val.data_err
                        metadata[meta_key]['var_info']['altitude'] = via = od()
                        
                        vi.update(val.var_info[var])
                        via.update(val.var_info['altitude'])
                    else:
                        add = 1
                        altitude = np.nan
                        data = val
                        if var in stat.data_err:
                            err = stat.err[var]
                        else:
                            err = np.nan
                    vi.update(stat.var_info[var])   
                    stop = idx + add
                    #check if size of data object needs to be extended
                    if stop >= data_obj._ROWNO:
                        #if totnum < data_obj._CHUNKSIZE, then the latter is used
                        data_obj.add_chunk(add)
                    
                    #write common meta info for this station
                    data_obj._data[idx:stop, 
                                   col_idx['latitude']] = stat['latitude']
                    data_obj._data[idx:stop, 
                                   col_idx['longitude']] = stat['longitude']
                    data_obj._data[idx:stop, 
                                   col_idx['altitude']] = stat['altitude']
                    data_obj._data[idx:stop, 
                                   col_idx['meta']] = meta_key
                                   
                    # write data to data object
                    data_obj._data[idx:stop, col_idx['time']] = time
                    data_obj._data[idx:stop, col_idx['stoptime']] = stat.stopdtime[0]
                    data_obj._data[idx:stop, col_idx['data']] = data
                    data_obj._data[idx:stop, col_idx['dataaltitude']] = altitude
                    data_obj._data[idx:stop, col_idx['varidx']] = var_idx
                    
                    if read_err:
                        data_obj._data[idx:stop, col_idx['dataerr']] = err
                    
                    if not var in meta_idx[meta_key]:
                        meta_idx[meta_key][var] = []
                    meta_idx[meta_key][var].extend(list(range(idx, stop)))
                    
                    if not var in metadata[meta_key]['variables']:
                        metadata[meta_key]['variables'].append(var)
                    
                    idx += add
                    
            except Exception as e:
                self.read_failed.append(_file)
                self.logger.exception('Failed to read file {} (ERR: {})'
                                      .format(os.path.basename(_file),
                                              repr(e)))
                
        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]
        #data_obj.data_revision[self.DATA_ID] = self.data_revision
        self.data = data_obj
        return data_obj
        
    def _get_exclude_filelist(self):
        """Get list of filenames that are supposed to be ignored"""
        exclude = []
        import glob
        files = glob.glob('{}/EXCLUDE/*.txt'.format(self.DATASET_PATH))
        for i, file in enumerate(files):
            if not os.path.basename(file) in self.EXCLUDE_CASES:
                continue
            count = 0
            num = None
            indata = False
            with open(file) as f:
                for line in f:
                    if indata:
                        exclude.append(line.strip())
                        count += 1
                    elif 'Number of' in line:
                        num = int(line.split(':')[1].strip())
                        indata = True
                    
            if not count == num:
                raise Exception
        self.exclude_files = list(dict.fromkeys(exclude))
        return self.exclude_files
    
    #TODO: check performance (it is usually slow...)
    def get_file_list(self, vars_to_retrieve=None, file_pattern=None):
        """Perform recusive file search for all input variables
        
        Note
        ----
        Overloaded implementation of base class, since for Earlinet, the 
        paths are variable dependent
        
        Parameters
        ----------
        vars_to_retrieve : list
            list of variables to retrieve
        
        Returns
        -------
        list
            list containing file paths
        """
        
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        exclude = self._get_exclude_filelist()
        const.print_log.info('Fetching EARLINET data files. '
                             'This might take a while...')
        patterns = []
        for var in vars_to_retrieve:
            if not var in self.VAR_PATTERNS_FILE:                
                from pyaerocom.exceptions import VarNotAvailableError
                raise VarNotAvailableError('Input variable {} is not supported'
                                           .format(var))
            
            pattern = self.VAR_PATTERNS_FILE[var]
            if file_pattern is not None:
                if '/' in pattern:
                    raise NotImplementedError('Cannot apply file pattern to '
                                              'wildcard path mask including / ')
                elif '.' in file_pattern:
                    raise NotImplementedError('filetype delimiter . not '
                                              'supported')
                spl = pattern.split('.')
                if not '*' in spl[0]:
                    raise AttributeError('Invalid file pattern: {}'.format(pattern))
                spl[0] = spl[0].replace('*', file_pattern)
                pattern = '.'.join(spl)
                
            patterns.append(pattern)
            
        matches = []
        for root, dirnames, files in os.walk(self.DATASET_PATH):
            paths = [os.path.join(root, f) for f in files]
            for pattern in patterns:
                for path in paths:
                    file = os.path.basename(path)
                    if not fnmatch.fnmatch(file, pattern):
                        continue
                    elif file in exclude:
                        self.excluded_files.append(path)
                    else:
                        matches.append(path)
        self.files = files = list(dict.fromkeys(matches))
        return files
    
    def copy(self):
        """Make and return a deepcopy of this object"""
        from copy import deepcopy
        return deepcopy(self)
    
if __name__=="__main__":
    import matplotlib.pyplot as plt

    
    plt.close('all')
    r = ReadEarlinet()
    
    print(r.DATASET_PATH)
    
    data = r.read(['ec532aer', 'bscatc532aer'])
    
    data._check_index()

# =============================================================================
#     data = read.read('ec532aer')
#     print(data)
#     
#     stat = data.to_station_data(0)
#     
#     
#     merged0 = data.to_station_data('Evora')
#     merged = data.to_station_data('Evora', freq='monthly')
#     
#     print(merged)
#     
#     merged.ec532aer.plot()
#     
#     arr0 = merged0.ec532aer
#     arr = merged.ec532aer
#     
# =============================================================================
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################
# readoddata.py
#
# observational data reading class
#
# this file is part of the aerocom_pt package
#
#################################################################
# Created 20171024 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2017 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no, jonas.gliss@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA
import os, logging

from pyaerocom.exceptions import NetworkNotImplemented, NetworkNotSupported
from pyaerocom.io.read_aeronet_sdav2 import ReadAeronetSdaV2
from pyaerocom.io.read_aeronet_sdav3 import ReadAeronetSdaV3
from pyaerocom.io.read_aeronet_invv2 import ReadAeronetInvV2
from pyaerocom.io.read_aeronet_invv3 import ReadAeronetInvV3
from pyaerocom.io.read_aeronet_sunv2 import ReadAeronetSunV2
from pyaerocom.io.read_aeronet_sunv3 import ReadAeronetSunV3
from pyaerocom.io.read_earlinet import ReadEarlinet
from pyaerocom.io.read_ebas import ReadEbas
from pyaerocom.io.read_aasetal import ReadSulphurAasEtAl
from pyaerocom.io.read_gaw import ReadGAW


from pyaerocom.io.cachehandler_ungridded import CacheHandlerUngridded
from pyaerocom.ungriddeddata import UngriddedData

from pyaerocom import const, print_log

# TODO Note: Removed infiles (list of files from which datasets were read, since it 
# was not used anywhere so far)
class ReadUngridded(object):
    """Factory class for reading of ungridded data based on obsnetwork ID"""
    SUPPORTED = [ReadAeronetInvV3,
                 ReadAeronetInvV2,
                 ReadAeronetSdaV2,
                 ReadAeronetSdaV3,
                 ReadAeronetSunV2,
                 ReadAeronetSunV3,
                 ReadEarlinet,
                 ReadEbas,
                 ReadGAW,
                 ReadSulphurAasEtAl]

    # when this file exists, an existing cache file is not read
    _DONOTCACHEFILE = None
    if isinstance(const._cachedir, str) and os.path.exists(const._cachedir):
        _DONOTCACHEFILE = os.path.join(const._cachedir, 'DONOTCACHE')

    def __init__(self, datasets_to_read=None,
                 vars_to_retrieve=None, ignore_cache=False):
        if datasets_to_read is None:
            datasets_to_read = const.AERONET_SUN_V3L2_AOD_DAILY_NAME
        #will be assigned in setter method of dataset_to_read
        self._datasets_to_read = []
        #: dictionary containing reading classes for each dataset to read (will
        #: be filled in setter of datasets_to_read)
        self._readers = {}
        
        self.datasets_to_read = datasets_to_read
    
        # optional: list of variables that are supposed to be imported, if 
        # None, all variables provided by the corresponding network are loaded
        self.vars_to_retrieve = vars_to_retrieve
        
        # initiate a logger for this class
        self.logger = logging.getLogger(__name__)
        
        if ignore_cache:
            self.logger.info('Deactivating caching')
            const.CACHING = False
           
    @property
    def DATASET_PATH(self):
        """Data directory of dataset to read 
        
        Raises exception if more than one dataset to read is specified
        """
        if not len(self._datasets_to_read) == 1:
            raise ValueError('Conflict: multiple datasets are assigned to be '
                             'read, but dataset path can only be retrieved for '
                             'single dataset retrievals')
        return self.get_reader(self._datasets_to_read[0]).DATASET_PATH
    
    @property
    def ignore_cache(self):
        if os.path.exists(self._DONOTCACHEFILE) or not const.CACHING:
            return True
        return False
    
    @property
    def dataset_to_read(self):
        """Helper that returns the dataset to be read
        
        Note
        ----
        Only works if a single dataset is assigned in :attr:`datasets_to_read`,
        else throws an ValueError.
        
        Raises
        ------
        ValueError
            if :attr:`datasets_to_read` contains no or more than one entry.
        """
        dsr = self.datasets_to_read
        if len(dsr) == 0:
            raise ValueError('Could not fetch reader class. No dataset '
                             'assigned in attr. datasets_to_read')
        elif len(dsr) > 1:
            raise ValueError('Could not fetch reader class. More than one '
                             'dataset is assigned in attr. '
                             'datasets_to_read')
        return dsr[0]
    
    @property
    def datasets_to_read(self):
        """List of datasets supposed to be read"""
        return self._datasets_to_read
    
    @datasets_to_read.setter
    def datasets_to_read(self, datasets):
        if isinstance(datasets, str):
            datasets = [datasets]
        elif not isinstance(datasets, (tuple, list)):
            raise IOError('Invalid input for parameter datasets_to_read')
        avail = []
        for ds in datasets:
            try:
                self.find_read_class(ds)
                avail.append(ds)
            except NetworkNotSupported:
                print_log.warning('Removing {} from list of datasets to read '
                                  'in ReadUngridded class. Reason: network '
                                  'not supported or data is not available'
                                  .format(ds))
        self._datasets_to_read = avail    
    
    def dataset_provides_variables(self, dataset_to_read=None):
        """List of variables provided by a certain dataset"""
        if dataset_to_read is None:
            dataset_to_read = self.dataset_to_read
        if dataset_to_read in self._readers:
            return self._readers[dataset_to_read].PROVIDES_VARIABLES
        return self.find_read_class(dataset_to_read).PROVIDES_VARIABLES
      
    def get_reader(self, dataset_to_read=None):
        """Helper method that returns loaded reader class
        
        Parameters
        -----------
        dataset_to_read : str
            Name of dataset
        
        Returns
        -------
        ReadUngriddedBase
            instance of reading class (needs to be implementation of base 
            class :class:`ReadUngriddedBase`)
        
        Raises
        ------
        NetworkNotSupported
            if network is not supported by pyaerocom
        NetworkNotImplemented
            if network is supported but no reading routine is implemented yet
        """
        if dataset_to_read is None:
            dataset_to_read = self.dataset_to_read
        elif not dataset_to_read in self.supported_datasets:
            raise NetworkNotSupported('Could not fetch reader class: Input '
                                      'network {} is not supported by '
                                      'ReadUngridded'.format(dataset_to_read))
        if not dataset_to_read in self._readers:
            self.find_read_class(dataset_to_read)
        return self._readers[dataset_to_read]
    
    def find_read_class(self, dataset_to_read):
        """Find reading class for dataset name
        
        Loops over all reading classes available in :attr:`SUPPORTED` and finds
        the first one that matches the input dataset name, by checking the 
        attribute :attr:`SUPPORTED_DATASETS` in each respective reading class
        
        Parameters
        -----------
        dataset_to_read : str
            Name of dataset
        
        Returns
        -------
        ReadUngriddedBase
            instance of reading class (needs to be implementation of base 
            class :class:`ReadUngriddedBase`)
        
        Raises
        ------
        NetworkNotSupported
            if network is not supported by pyaerocom
        NetworkNotImplemented
            if network is supported but no reading routine is implemented yet
        
        """
        if dataset_to_read in self._readers:
            return self._readers[dataset_to_read]
        if not dataset_to_read in const.OBS_IDS:
            raise NetworkNotSupported("Network {} is not supported or data is "
                                      "not available on this machine"
                                      .format(dataset_to_read))
        for cls in self.SUPPORTED:
            if dataset_to_read in cls.SUPPORTED_DATASETS:
                self._readers[dataset_to_read] = cls(dataset_to_read)
                return self._readers[dataset_to_read]
        raise NetworkNotImplemented("No reading class available yet for dataset "
                                    "{}".format(dataset_to_read))
        
    
    def read_dataset(self, dataset_to_read, vars_to_retrieve=None, 
                     **kwargs):
        """Read dataset into an instance of :class:`ReadUngridded`
        
        Note
        ----
        This method does not assign loaded data obj to class attribute 
        :attr:`data` (only :func:`read` does)
        
        Parameters
        ----------
        dataset_to_read : str
            name of dataset
        vars_to_retrieve : list
            list of variables to be retrieved. If None (default), the default
            variables of each reading routine are imported
            
        Returns
        --------
        UngriddedData
            data object
        """
        _caching = None
        if len(kwargs) > 0:
            _caching = const.CACHING 
            const.CACHING = False
            
            print_log.info('Received additional reading constraints, '
                           'ignoring caching')
        if vars_to_retrieve is None:
            # Note: self.vars_to_retrieve may be None as well, then
            # default variables of each network are read
            vars_to_retrieve = self.vars_to_retrieve 
        
        reader = self.get_reader(dataset_to_read)
        
        if vars_to_retrieve is None:
            vars_to_retrieve = reader.PROVIDES_VARIABLES
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
            
        # Since this interface enables to load multiple datasets, each of 
        # which support a number of variables, here, only the variables are 
        # considered that are supported by the dataset
        vars_available = [var for var in vars_to_retrieve if var in 
                          reader.PROVIDES_VARIABLES]
        
        cache = CacheHandlerUngridded(reader)
        if not self.ignore_cache:
            # initate cache handler    
            for var in vars_available:
                try:
                    cache.check_and_load(var_name=var)
                except:
                    self.logger.exception('Fatal: compatibility error between '
                                          'old cache file {} and current version '
                                          'of code ')
                    
            
        vars_to_read = [v for v in vars_available if not v in cache.loaded_data]
        data_read = None
        if len(vars_to_read) > 0:
            
            _loglevel = print_log.level
            print_log.setLevel(logging.INFO)
            data_read = reader.read(vars_to_read, **kwargs)
            print_log.setLevel(_loglevel)
            
            for var in vars_to_read:
                # write the cache file
                if not self.ignore_cache:
                    try:
                        cache.write(data_read, var)
                    except Exception as e:
                        _caching = False
                        print_log.warning('Failed to write to cache directory. '
                                          'Error: {}. Deactivating caching in '
                                          'pyaerocom'.format(repr(e)))
        
        if len(vars_to_read) == len(vars_available):
            data_out = data_read
        else:
            data_out = UngriddedData()
            for var in vars_available:
                if var in cache.loaded_data:
                    data_out.append(cache.loaded_data[var])
            if data_read is not None:
                data_out.append(data_read)
                
        if _caching is not None:
            const.CACHING = _caching
        return data_out
    
    def read(self, datasets_to_read=None, vars_to_retrieve=None, **kwargs):
        """Read observations

        Iter over all datasets in :attr:`datasets_to_read`, call 
        :func:`read_dataset` and append to data object
        
        Example
        -------
        >>> import pyaerocom.io.readungridded as pio
        >>> from pyaerocom import const
        >>> obj = pio.ReadUngridded(dataset_to_read=const.AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME)
        >>> obj.read()
        >>> print(obj)
        >>> print(obj.metadata[0.]['latitude'])
        """
        if datasets_to_read is not None:
            self.datasets_to_read = datasets_to_read
        if vars_to_retrieve is not None:
            self.vars_to_retrieve = vars_to_retrieve
            
        data = UngriddedData()
        for ds in self.datasets_to_read:
            self.logger.info('Reading {} data'.format(ds))
            data.append(self.read_dataset(ds, vars_to_retrieve, **kwargs))
            self.logger.info('Successfully imported {} data'.format(ds))
        return data
    
    @property
    def SUPPORTED_DATASETS(self):
        """Returns list of strings containing all supported dataset names"""
        l = []
        for r in self.SUPPORTED:
            l.extend(r.SUPPORTED_DATASETS)
        return l 
    
    @property
    def supported_datasets(self):
        return self.SUPPORTED_DATASETS
    
    def __str__(self):
        #raise NotImplementedError("Requires review after API changes")
        s=''
        for ds in self.datasets_to_read:
            s += '\n{}'.format(self.get_reader(ds))
        return s
    
    # OUTSOURCED
    def read_datasetOLD(self, dataset_to_read, vars_to_retrieve=None, 
                     **kwargs):
        """Read single dataset into instance of :class:`ReadUngridded`
        
        Note
        ----
        This method does not write class attribute :attr:`data` (only
        :func:`read` does)
        
        Parameters
        ----------
        dataset_to_read : str
            name of dataset
        vars_to_retrieve : list
            list of variables to be retrieved. If None (default), the default
            variables of each reading routine are imported
            
        Returns
        --------
        UngriddedData
            data object
        """
        _caching = None
        if len(kwargs) > 0:
            _caching = const.CACHING 
            const.CACHING = False
            
            print_log.info('Received additional reading constraints, '
                           'ignoring caching')
        if vars_to_retrieve is None:
            # Note: self.vars_to_retrieve may be None as well, then
            # default variables of each network are read
            vars_to_retrieve = self.vars_to_retrieve 
            
        reader = self.get_reader(dataset_to_read)
        
        if vars_to_retrieve is None:
            vars_to_retrieve = reader.PROVIDES_VARIABLES
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
            
        # Since this interface enables to load multiple datasets, each of 
        # which support a number of variables, here, only the variables are 
        # considered that are supported by the dataset
        vars_available = [var for var in vars_to_retrieve if var in 
                          reader.PROVIDES_VARIABLES]
        
        # read the data sets
        cache_hit_flag = False
        
        if not self.ignore_cache:
            # initate cache handler
            try:
                cache = CacheHandlerUngridded(reader, vars_available, **kwargs)
                if cache.check_and_load():
                    all_avail = True
                    for var in vars_available:
                        if not var in cache.loaded_data:
                            all_avail = False
                            break
                    if all_avail:
                        print_log.info('Found Cache match for {}'
                                       .format(dataset_to_read))
                        cache_hit_flag = True
                        data = cache.loaded_data
            except:
                self.logger.exception('Fatal: compatibility error between old '
                                      'cache file and current version of code ')
                cache_hit_flag = False
            
        if not cache_hit_flag:
            print_log.info('No Cache match found for {} in {}. '
                           'Reading from files (this '
                           'may take a while)'.format(dataset_to_read,
                                                      const.CACHEDIR))
            _loglevel = print_log.level
            print_log.setLevel(logging.INFO)
            data = reader.read(vars_available, **kwargs)
            print_log.setLevel(_loglevel)
        
        self.revision[dataset_to_read] = reader.data_revision
        self.data_version[dataset_to_read] = reader.__version__
        
        # write the cache file
        if not cache_hit_flag and not self.ignore_cache:
            try:
                cache.write(data)
            except Exception as e:
                _caching = False
                print_log.warning('Failed to write to cache directory:\n{}.\n'
                               'Deactivating caching in pyaerocom'.format(repr(e)))
        
        if _caching is not None:
            const.CACHING = _caching
        return data
if __name__=="__main__":

    reader = ReadUngridded()
    
    data = reader.read('EBASMC', 
                       vars_to_retrieve=['conctc', 'concoa', 'concbc', 'conceqbc'])
    
    
# =============================================================================
#     data = reader.read([const.AERONET_SUN_V2L2_AOD_DAILY_NAME,
#                         const.AERONET_SUN_V3L2_AOD_DAILY_NAME], 
#                     vars_to_retrieve='od550aer', 
#                     last_file=10)
#     
# =============================================================================


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains scatter plot routines for Aerocom data.
"""

from pyaerocom import const, BASEMAP_AVAILABLE
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import sys
import cartopy.crs as ccrs
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter

# TODO: Not used currently, required??
# text positions for the annotations
XYPOS=[]
XYPOS.append((.01, 0.95))
XYPOS.append((0.01, 0.90))
XYPOS.append((0.3, 0.90))
XYPOS.append((0.01, 0.86))
XYPOS.append((0.3, 0.86))
XYPOS.append((0.01, 0.82))
XYPOS.append((0.3, 0.82))
XYPOS.append((0.01, 0.78))
XYPOS.append((0.3, 0.78))
XYPOS.append((0.8, 0.1))
XYPOS.append((0.8, 0.06))


def plotsitelocation(model_name, model_data=None, obs_data=None, options=None, 
                     verbose=True):
    """method to plot scatterplots"""

    plt_name = 'SITELOCATION'
    var_to_run = options['VariablesToRun'][0]
    obs_network_name = options['ObsNetworkName'][0]
    obs_data_as_series = obs_data.to_timeseries(start_date=options['StartDate'], 
                                                end_date=options['EndDate'], 
                                                freq='D')
    obs_lats = [obs_data_as_series[i]['latitude'] 
                for i in range(len(obs_data_as_series))]
    obs_lons = [obs_data_as_series[i]['longitude'] 
                for i in range(len(obs_data_as_series))]
    obs_names = [obs_data_as_series[i]['station_name']
                 for i in range(len(obs_data_as_series))]
    # model_station_data = model_data.interpolate([("latitude", obs_lats), ("longitude", obs_lons)])
    # times_as_dt64 = pa.helpers.cftime_to_datetime64(model_station_data.time)
    # model_data_as_series = pa.helpers.to_time_series_griesie(model_station_data.grid.data, obs_lats, obs_lons,
    #                                                          times_as_dt64, var_name = [var_to_run])

    model_data_as_series = model_data.to_time_series([("latitude", obs_lats), 
                                                      ("longitude", obs_lons)])

    df_time = pd.DataFrame()
    df_points = pd.DataFrame()
    station_no = 0
    for i in range(len(obs_data_as_series)):
        _len = len(obs_data_as_series[i][var_to_run])
        # print('{} length: {}'.format(obs_names[i],_len))
        if _len > 0:
            _nansum = np.nansum(obs_data_as_series[i][var_to_run])
            # _isnan = np.isnan(_nansum)
            # print('{} nansum: {:.3f}'.format(obs_names[i],np.nansum(obs_data_as_series[i][var_to_run])))
            # print('{} isnan: {}'.format(obs_names[i],_isnan))
            if _nansum > np.float_(0.):
                station_no += 1
                # print('{} station_no: {}'.format(obs_names[i],station_no))
            else:
                print('{} removed due to NaNs only'.format(obs_names[i]))
        else:
            continue
        # put obs and model in DataFrame to make them use the same time index
        df_time_temp = pd.DataFrame(obs_data_as_series[i][var_to_run],
                                    columns=[obs_network_name])
        df_points = df_points.append(df_time_temp)
        # df_time_temp[model_name] = model_data_as_series[i][var_to_run]*1.E3
        df_time_temp[model_name] = ( model_data_as_series[i][var_to_run] *
                                const.VARS[var_to_run]['scat_scale_factor'])
        # df_time has now all time steps where either one of the obs or model data have data
        #
        # df_points = df_points.append(pd.DataFrame(np.float_(df_time_temp.values), columns=df_time_temp.columns))
        df_time = df_time.append(pd.DataFrame(df_time_temp, 
                                              columns=df_time_temp.columns))

    # remove all indices where either one of the data pairs is NaN
    # mainly done to get the number of days right.
    # df_time.corr() gets it right without
    df_time = df_time.dropna(axis=0, how='any')
    df_points = df_points.dropna()
    print('# of measurements: {}'.format(len(df_points)))

    filter_name = 'WORLD-wMOUNTAINS'
    filter_name = 'WORLD'
    time_step_name = 'mALLYEARdaily'
    # OD550_AER_an2008_YEARLY_WORLD_SCATTERLOG_AeronetSunV3Lev2.0.daily.ps.png
    # if df_time[model_name].index[0].year != df_time[model_name].index[-1].year:
    years_covered = df_time[model_name].index[:].year.unique().sort_values()
    if len(years_covered) > 1:
        figname = '{}_{}_an{}-{}_{}_{}_{}_{}.png'.format(model_name,
                                                         var_to_run,years_covered[0],
                                                      years_covered[-1],
                                                      time_step_name, filter_name, plt_name,
                                                      obs_network_name)
        plotname = "{}-{} {}".format(years_covered[0], years_covered[-1], 'daily')
        title = "{} {} station list {}-{}".format(var_to_run, filter_name, years_covered[0], years_covered[-1])
    else:
        figname = '{}_{}_an{}_{}_{}_{}_{}.png'.format(model_name,
                                                      var_to_run,years_covered[0],
                                                      time_step_name, filter_name, plt_name,
                                                      obs_network_name)
        plotname = "{} {}".format(years_covered[0], 'daily')
        title = "{} {} station list {}".format(var_to_run, filter_name, years_covered[0])

    if verbose:
        sys.stdout.write(figname+"\n")

    lat_low = -90
    lat_high = 90.
    lon_low = -180.
    lon_high = 180.
    
    # TODO: review basemap dependency
    basemap_flag=False
    if basemap_flag and BASEMAP_AVAILABLE:
        m = Basemap(projection='cyl', llcrnrlat=lat_low, urcrnrlat=lat_high,
                    llcrnrlon=lon_low, urcrnrlon=lon_high, resolution='c', fix_aspect=False)

        x, y = m(obs_lons, obs_lats)
        # m.drawmapboundary(fill_color='#99ffff')
        # m.fillcontinents(color='#cc9966', lake_color='#99ffff')
        plot = m.scatter(x, y, 4, marker='o', color='r', )
        m.drawmeridians(np.arange(-180,220,40),labels=[0,0,0,1], fontsize=10)
        m.drawparallels(np.arange(-90,120,30),labels=[1,1,0,0], fontsize=10)
        # axis = plt.axis([LatsToPlot.min(), LatsToPlot.max(), LonsToPlot.min(), LonsToPlot.max()])
        ax = plot.axes
        m.drawcoastlines()
    else:
        ax = plt.axes([0.15,0.1,0.8,0.8],projection=ccrs.PlateCarree())
        ax.set_ylim([lat_low, lat_high])
        ax.set_xlim([lon_low, lon_high])
        #ax.set_aspect(2)

        ax.coastlines()
        plot = plt.scatter(obs_lons, obs_lats, 8, marker='o', color='r')
        #plot.axes.set_aspect(1.8)


        # lon_formatter = LongitudeFormatter(number_format='.1f', degree_symbol='')
        # lat_formatter = LatitudeFormatter(number_format='.1f', degree_symbol='')
        # ax.xaxis.set_major_formatter(lon_formatter)
        # ax.yaxis.set_major_formatter(lat_formatter)
        xticks = ax.set_xticks([-180., -120., -60., 0., 60, 120, 180])
        yticks = ax.set_yticks([-90., -60, -30, 0., 30, 60, 90])
        # ax.annotate('source: AEROCOM', xy=(0.93, 0.04), xycoords='figure fraction',
        #             horizontalalignment='right', fontsize=9, bbox=dict(boxstyle='square', facecolor='none',
        #                                                                 edgecolor='black'))
        ax.annotate('longitude', xy=(0.55, 0.12), xycoords='figure fraction',
                    horizontalalignment='center', fontsize=12, )
        ax.annotate('latitude', xy=(0.07, 0.55), xycoords='figure fraction', rotation=90,
                    horizontalalignment='center', fontsize=12, )
        # ax.set_xlabel = 'longitude'
        # ax.set_ylabel = 'latitude'
        ax.annotate(model_name, xy=(174., -83.), xycoords='data', horizontalalignment='right', fontsize=13,
                    fontweight='bold',
                    color='black', bbox=dict(boxstyle='square', facecolor='white', edgecolor='none', alpha=0.7))
        ax.annotate("No of stations: {}".format(station_no), xy=(-174., -83.), xycoords='data',
                    fontweight='bold',
                    horizontalalignment='left', fontsize=13,
                    color='black', bbox=dict(boxstyle='square', facecolor='white', edgecolor='none', alpha=0.7))



    plt.title(title, fontsize=13)
    plt.xticks(fontsize=11)
    plt.yticks(fontsize=11)
    plt.xlabel = 'longitude'
    plt.ylabel = 'latitude'



    plt.savefig(figname, dpi=300)
    plt.close()
